
_missing = object()


class cached_property(property):  # pragma: no cover
    """

    Import from:
        https://github.com/pallets/werkzeug/blob/master/werkzeug/utils.py
    """

    # implementation detail: A subclass of python's builtin property
    # decorator, we override __get__ to check for a cached value. If one
    # choses to invoke __get__ by hand the property will still work as
    # expected because the lookup logic is replicated in __get__ for
    # manual invocation.

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __set__(self, obj, value):
        obj.__dict__[self.__name__] = value

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, _missing)
        if value is _missing:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value



""":mod:`ring.wire` --- Universal method/function wrapper.
==========================================================
"""
import six
from .callable import Callable
from .wire import descriptor_bind
from ._compat import functools


class RopeCore(object):

    def __init__(self, callable, rope):
        super(RopeCore, self).__init__()
        self.callable = callable
        self.rope = rope
        self.wire_class = rope.wire_class


class MethodRopeMixin(object):

    def __init__(self, *args, **kwargs):
        super(MethodRopeMixin, self).__init__(*args, **kwargs)
        assert not self.callable.is_barefunction

    def __get__(self, obj, type=None):
        cw = self.callable
        co = cw.wrapped_object
        owner = descriptor_bind(co, obj, type)
        if owner is None:  # invalid binding but still wire it
            owner = obj if obj is not None else type
        wrapper_name_parts = ['__wire_', cw.wrapped_callable.__name__]
        if owner is type:
            wrapper_name_parts.extend(('_', type.__name__))
        wrapper_name = ''.join(wrapper_name_parts)
        wrapper = getattr(owner, wrapper_name, None)
        if wrapper is None:
            boundmethod = co.__get__(obj, type)
            wire = self.wire_class(self, (obj, type))
            wrapper = functools.wraps(boundmethod)(wire)
            setattr(owner, wrapper_name, wrapper)
        return wrapper


class FunctionRopeMixin(object):

    def __init__(self, *args, **kwargs):
        super(FunctionRopeMixin, self).__init__(*args, **kwargs)
        assert self.callable.is_barefunction
        boundmethod = self.callable.wrapped_object
        wire = self.wire_class(self, None)
        self._wire = functools.wraps(boundmethod)(wire)

    def __getattr__(self, name):
        try:
            return self.__getattribute__(name)
        except AttributeError:
            pass
        return getattr(self._wire, name)


class CallableRopeMixin(object):

    def __init__(self, *args, **kwargs):
        super(CallableRopeMixin, self).__init__(*args, **kwargs)
        self.__call__ = functools.wraps(self.callable.wrapped_object)(self)

    def __call__(self, *args, **kwargs):
        return self._wire(*args, **kwargs)


class WireRope(object):

    def __init__(self, wire_class, core_class=RopeCore):
        self.wire_class = wire_class
        self.method_rope = type(
            '_MethodRope', (MethodRopeMixin, core_class), {})
        self.function_rope = type(
            '_FunctionRope', (FunctionRopeMixin, core_class), {})
        self.callable_function_rope = type(
            '_CallableFunctionRope',
            (CallableRopeMixin, FunctionRopeMixin, core_class), {})

    def __call__(self, function):
        """Wrap a function/method definition.

        :return: Wrapper object. The return type is up to given callable is
                 function or method.
        """
        wrapper = Callable(function)
        if wrapper.is_barefunction:
            rope_class = self.callable_function_rope
            wire_class_call = self.wire_class.__call__
            if six.PY3:
                if wire_class_call.__qualname__ == 'type.__call__':
                    rope_class = self.function_rope
            else:
                # method-wrapper test for CPython2.7
                # im_class == type test for PyPy2.7
                if type(wire_class_call).__name__ == 'method-wrapper' or \
                        wire_class_call.im_class == type:
                    rope_class = self.function_rope
        else:
            rope_class = self.method_rope
        rope = rope_class(wrapper, rope=self)
        return rope



from __future__ import absolute_import

import types
import six
from ._util import cached_property
from ._compat import inspect

__all__ = ('Callable', )


_inspect_iscoroutinefunction = getattr(
    inspect, 'iscoroutinefunction', lambda f: False)


class Callable(object):
    """A wrapper object including more information of callables."""

    def __init__(self, f):
        self.wrapped_object = f
        if not callable(f):
            f = f.__func__
        self.wrapped_callable = f
        self.is_wrapped_coroutine = getattr(f, '_is_coroutine', None)
        self.is_coroutine = self.is_wrapped_coroutine or \
            _inspect_iscoroutinefunction(f)

    @cached_property
    def signature(self):
        return inspect.signature(self.wrapped_callable)

    @cached_property
    def parameters(self):
        return list(self.signature.parameters.values())

    @cached_property
    def is_descriptor(self):
        return type(self.wrapped_object).__get__ is not types.FunctionType.__get__  # noqa

    @cached_property
    def is_barefunction(self):
        cc = self.wrapped_callable
        if six.PY34:
            method_name = cc.__qualname__.split('<locals>.')[-1]
            if method_name == cc.__name__:
                return True
            return False
        else:
            if self.is_descriptor:
                return False
            # im_class does not exist at this point
            return not (self.is_membermethod or self.is_classmethod)

    @cached_property
    def is_membermethod(self):
        """Test given argument is a method or not.

        :param ring.callable.Callable c: A callable object.
        :rtype: bool

        :note: The test is not based on python state but based on parameter
            name `self`. The test result might be wrong.
        """
        if six.PY34:
            if self.is_barefunction:
                return False
            if not self.is_descriptor:
                return True

        return self.first_parameter is not None \
            and self.first_parameter.name == 'self'

    @cached_property
    def is_classmethod(self):
        """Test given argument is a classmethod or not.

        :param ring.callable.Callable c: A callable object.
        :rtype: bool
        """
        if isinstance(self.wrapped_object, classmethod):
            return True
        if six.PY34:
            if self.is_barefunction:
                return False
            if not self.is_descriptor:
                return False

        return self.first_parameter is not None \
            and self.first_parameter.name == 'cls'

    @cached_property
    def first_parameter(self):
        parameters = self.parameters
        if not parameters:
            return None
        return parameters[0]



from __future__ import absolute_import

import six

import functools
if not hasattr(functools, 'singledispatch'):
    import singledispatch
    functools.singledispatch = singledispatch.singledispatch

if six.PY3:
    import inspect
else:
    import inspect2 as inspect  # noqa



from __future__ import with_statement

from setuptools import setup
import sys


def get_version():
    with open('wirerope/__version__.py') as f:
        empty, version = f.read().split('__version__ = ')
    assert empty == ''
    version = version.strip().strip("'")
    assert version.startswith('0.')
    return version


install_requires = [
    'six>=1.11.0',
]
tests_require = [
    'pytest>=3.0.2', 'pytest-cov',
]
docs_require = [
    'sphinx',
]

dev_require = tests_require + docs_require

# backports - py2
if sys.version_info[0] == 2:
    install_requires.extend([
        'inspect2>=0.1.0',
    ])

# backports - py34
if sys.version_info[:2] < (3, 4):
    install_requires.extend([
        'singledispatch>=3.4.0.3',
    ])


def get_readme():
    try:
        with open('README.rst') as f:
            return f.read().strip()
    except IOError:
        return ''


setup(
    name='wirerope',
    version=get_version(),
    description='Turn functions and methods into fully controllable objects',
    long_description=get_readme(),
    author='Jeong YunWon',
    author_email='wirerope@youknowone.org',
    url='https://github.com/youknowone/wirerope',
    packages=(
        'wirerope',
    ),
    package_data={},
    install_requires=install_requires,
    tests_require=tests_require + ['tox', 'tox-pyenv'],
    extras_require={
        'tests': tests_require,
        'docs': docs_require,
        'dev': dev_require,
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)  # noqa



""":mod:`ring.wire` --- Universal method/function wrapper.
==========================================================
"""
from ._compat import functools
from ._util import cached_property


@functools.singledispatch
def descriptor_bind(descriptor, obj, type):
    return obj


@descriptor_bind.register(staticmethod)
@descriptor_bind.register(classmethod)
def descriptor_bind_(descriptor, obj, type):
    return type


class Wire(object):
    """The core data object for each function for bound method.

    Inherit this class to implement your own Wire classes.

    - For normal functions, each function is directly wrapped by **Wire**.
    - For any methods or descriptors (including classmethod, staticmethod),
      each one is wrapped by :class:`ring.wire.MethodRopeMixin`
      and it creates **Wire** object for each bound object.
    """

    def __init__(self, rope, binding):
        self._rope = rope
        self._callable = rope.callable
        self._binding = binding
        if binding:
            self.__func__ = self._callable.wrapped_object.__get__(*binding)
        else:
            self.__func__ = self._callable.wrapped_object

    @cached_property
    def _bound_objects(self):
        if self._binding is None:
            return ()
        else:
            return (descriptor_bind(
                self._callable.wrapped_object, *self._binding),)
